@@ -19,7 +19,7 @@ from __future__ import print_function
 from __future__ import unicode_literals
 from __future__ import division
 from __future__ import absolute_import
-#from pynlpl.common import u
+from pynlpl.common import isstring
 import sys
 if sys.version < '3':
     from codecs import getwriter
@@ -33,16 +33,58 @@ import unicodedata
 import string
 import io
 import array
+import re
 from itertools import permutations
 from pynlpl.statistics import FrequencyList
 from pynlpl.datatypes import intarraytobytearray, bytearraytoint, containsnullbyte
 
-            
+
+WHITESPACE = [" ", "\t", "\n", "\r","\v","\f"]
+EOSMARKERS = ('.','?','!','。',';','؟','｡','？','！','।','։','՞','።','᙮','។','៕')
+REGEXP_URL = re.compile(r"^(?:(?:https?):(?:(?://)|(?:\\\\))|www\.)(?:[\w\d:#@%/;$()~_?\+-=\\\.&](?:#!)?)*")
+REGEXP_MAIL = re.compile(r"^[A-Za-z0-9\.\+_-]+@[A-Za-z0-9\._-]+(?:\.[a-zA-Z]+)+") #email
+TOKENIZERRULES = (REGEXP_URL, REGEXP_MAIL)
+
             
 class Windower(object):
-    """Moves a sliding window over a list of tokens, returning all ngrams"""
+    """Moves a sliding window over a list of tokens, upon iteration in yields all n-grams of specified size in a tuple.
+
+    Example without markers:
+    
+    >>> for ngram in Windower("This is a test .",3, None, None):
+    ...     print(" ".join(ngram))
+    This is a
+    is a test
+    a test .
+    
+    Example with default markers:
+        
+    >>> for ngram in Windower("This is a test .",3):    
+    ...     print(" ".join(ngram))
+    <begin> <begin> This
+    <begin> This is
+    This is a
+    is a test
+    a test .
+    test . <end>
+    . <end> <end>
+    """
 
     def __init__(self, tokens, n=1, beginmarker = "<begin>", endmarker = "<end>"):
+        """
+        Constructor for Windower
+        
+        :param tokens: The tokens to iterate over. Should be an itereable. Strings will be split on spaces automatically.
+        :type tokens: iterable
+        :param n: The size of the n-grams to extract
+        :type n: integer
+        :param beginmarker: The marker for the beginning of the sentence, defaults to "<begin>". Set to None if no markers are desired.
+        :type beginmarker: string or None
+        :param endmarker: The marker for the end of the sentence, defaults to "<end>". Set to None if no markers are desired.
+        :type endmarker: string or None      
+        """        
+        
+        
         if isinstance(tokens, str) or (sys.version < '3' and isinstance(tokens, unicode)):
             self.tokens = tuple(tokens.split())
         else:
@@ -52,7 +94,35 @@ class Windower(object):
         self.beginmarker = beginmarker
         self.endmarker = endmarker
 
+    def __len__(self):
+        """Returns the number of n-grams in the data (quick computation without iteration)
+
+        Without markers:
+            
+        >>> len(Windower("This is a test .",3, None, None))                   
+        3
+        
+        >>> len(Windower("This is a test .",2, None, None))
+        4
+        
+        >>> len(Windower("This is a test .",1, None, None))                   
+        5
+                
+        With default markers:
+        
+        >>> len(Windower("This is a test .",3))                
+        7
+        
+        """        
+
+        c = (len(self.tokens) - self.n) + 1
+        if self.beginmarker: c += self.n-1
+        if self.endmarker: c += self.n-1        
+        return c
+
+
     def __iter__(self):
+        """Yields an n-gram (tuple) at each iteration"""
         l = len(self.tokens)
 
         if self.beginmarker:
@@ -190,31 +260,155 @@ def calculate_overlap(haystack, needle, allowpartial=True):
                 solutions.append( (needle, 0) )
 
     return solutions
+
+
+
             
-
-def crude_tokenizer(line):    
-    return tokenise(line) #backwards-compatibility
-
-def tokenize(line):
-    return tokenise(line)
-
-def tokenise(line):
-    """A simple tokeniser"""
-    tokens = []
-    buffer = ''
-    for c in line.strip():
-        if c == ' ' or c in string.punctuation:
-            if buffer:
-                tokens.append(buffer)
-                buffer = ''
-            if c in string.punctuation:
-                tokens.append(c)
-        else:
-            buffer += c          
-    if buffer: 
-        tokens.append(buffer)  
-    return tokens
+class Tokenizer(object):    
+    """A tokenizer and sentence splitter, which acts on a file/stream-like object and when iterating over the object it yields
+    a lists of tokens (in case the sentence splitter is active (default)), or a token (if the sentence splitter is deactivated).
+    """
     
+    def __init__(self, stream, splitsentences=True, onesentenceperline=False, regexps=TOKENIZERRULES):
+        """
+        Constructor for Tokenizer
+        
+        :param stream: An iterable or file-object containing the data to tokenize
+        :type stream: iterable or file-like object
+        :param splitsentences: Enable sentence splitter? (default=_True_)
+        :type splitsentences: bool
+        :param onesentenceperline: Assume input has one sentence per line? (default=_False_)
+        :type onesentenceperline: bool
+        :param regexps: Regular expressions to use as tokeniser rules in tokenisation (default=_pynlpl.textprocessors.TOKENIZERRULES_) 
+        :type regexps:  Tuple/list of regular expressions to use in tokenisation     
+        """
+                      
+        self.stream = stream
+        self.regexps = regexps
+        self.splitsentences=splitsentences
+        self.onesentenceperline = onesentenceperline
+        
+    def __iter__(self):
+        for line in self.stream:
+            line = line.strip()
+            if line:
+                if buffer: buffer += "\n"
+                buffer += line
+                        
+            if (self.onesentenceperline or not line) and buffer: 
+                if self.splitsentences:
+                    yield split_sentences(tokenize(buffer))
+                else:            
+                    for token in tokenize(buffer, self.regexps):
+                        yield token
+    
+        if buffer:        
+            if self.splitsentences:
+                yield split_sentences(tokenize(buffer))
+            else:            
+                for token in tokenize(buffer, self.regexps):
+                    yield token
+    
+            
+                            
+    
+def tokenize(text, regexps=TOKENIZERRULES):
+    """Tokenizes a string and returns a list of tokens
+    
+    :param text: The text to tokenise
+    :type text: string 
+    :param regexps: Regular expressions to use as tokeniser rules in tokenisation (default=_pynlpl.textprocessors.TOKENIZERRULES_) 
+    :type regexps:  Tuple/list of regular expressions to use in tokenisation
+    :rtype: Returns a list of tokens 
+
+    Examples:
+       
+    >>> for token in tokenize("This is a test."):
+    ...    print(token)
+    This
+    is
+    a
+    test
+    .
+          
+       
+    """
+    
+    for i,regexp in list(enumerate(regexps)):
+        if isstring(regexp):
+            regexps[i] = re.compile(regexp)
+    
+    tokens = []
+    begin = 0
+    for i, c in enumerate(text):
+        if begin > i: 
+            continue
+        elif i == begin:
+            m = False
+            for regexp in regexps:
+                m = regexp.findall(text[i:i+300])
+                if m:
+                    tokens.append(m[0])
+                    begin = i + len(m[0])
+                    break
+            if m: continue
+        
+        if c in string.punctuation or c in WHITESPACE:
+            prev = text[i-1] if i > 0 else ""
+            next = text[i+1] if i < len(text)-1 else ""
+            
+            if (c == '.' or c == ',') and prev.isdigit() and next.isdigit():
+                #punctuation in between numbers, keep as one token
+                pass
+            elif (c == "'" or c == "`") and prev.isalpha() and next.isalpha():
+                #quote in between chars, keep...
+                pass    
+            elif c not in WHITESPACE and next == c: #group clusters of identical punctuation together
+                continue
+            elif c == '\r' and prev == '\n':
+                #ignore
+                begin = i+1
+                continue
+            else:
+                token = text[begin:i]
+                if token: tokens.append(token)
+                    
+                if c not in WHITESPACE:
+                    tokens.append(c) #anything but spaces and newlines (i.e. punctuation) counts as a token too
+                begin = i + 1 #set the begin cursor
+    
+    if begin <= len(text) - 1:
+        token = text[begin:]
+        tokens.append(token)                    
+
+    return tokens
+                    
+
+def crude_tokenizer(text):    
+    """Replaced by tokenize(). Alias"""
+    return tokenize(text) #backwards-compatibility, not so crude anymore
+
+def tokenise(text, regexps=TOKENIZERRULES): #for the British
+    """Alias for the British"""
+    return tokenize(text)
+
+def is_end_of_sentence(tokens,i ):
+    # is this an end-of-sentence marker? ... and is this either 
+    # the last token or the next token is NOT an end of sentence 
+    # marker as well? (to deal with ellipsis etc)
+    return tokens[i] in EOSMARKERS and (i == len(tokens) - 1 or not tokens[i+1] in EOSMARKERS)
+
+def split_sentences(tokens):
+    """Split sentences (based on tokenised data), returns sentences as a list of lists of tokens, each sentence is a list of tokens"""
+    begin = 0
+    for i, token in enumerate(tokens):
+        if is_end_of_sentence(tokens, i): 
+            yield tokens[begin:i+1]
+            begin = i+1
+    if begin <= len(tokens)-1:    
+        yield tokens[begin:]
+
+
 
 def strip_accents(s, encoding= 'utf-8'):
     """Strip characters with diacritics and return a flat ascii representation"""
@@ -245,6 +439,19 @@ def swap(tokens, maxdist=2):
             break
         
 
+def find_keyword_in_context(tokens, keyword, contextsize=1):
+    """Find a keyword in a particular sequence of tokens, and return the local context. Contextsize is the number of words to the left and right. The keyword may have multiple word, in which case it should to passed as a tuple or list"""
+    if isinstance(keyword,tuple) and isinstance(keyword,list):
+        l = len(keyword)
+    else:
+        keyword = (keyword,)
+        l = 1
+    n = l + contextsize*2
+    focuspos = contextsize + 1
+    for ngram in Windower(tokens,n,None,None): 
+        if ngram[focuspos:focuspos+l] == keyword:
+            yield ngram[:focuspos], ngram[focuspos:focuspos+l],ngram[focuspos+l+1:]
+
 
 class Classer(object):
     """The Classer can encode and decode tokens to an integer representation. It is constructed using a frequency list."""
@@ -398,5 +605,7 @@ class Classer(object):
             
 
     
-    
+if __name__ == "__main__":    
+    import doctest
+    doctest.testmod()
         
